/******************************************************************************
* This is a very simple implementation of grep, it uses the reg_exp_class
* to search for strings in files.
* This source code is copyright by bertel@post4.tele.dk
* Permission is granted to anyone to use and redistribute it freely.
*
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "regexp.h"
#include <ctype.h>

#ifndef __GNUC__
#include <io.h>
#include <direct.h>

#endif

void help(BOOL regex);
void grep(const char *file);
char line[0x4000];
char l_line[0x4000];
reg_exp_class reg_exp;
char pattern[1024];

BOOL ignore_case;
BOOL use_regex;
BOOL got_pattern;
BOOL show_help;
BOOL read_a_file;

#ifndef __GNUC__
void expand(char *file, char *path);
void parse_path(char *file, char *path);
BOOL parse_subdir = FALSE;
#endif

int main(int argc, char *args[])
{
  int i, n;

  for(i = 0; i < argc; i++)
  {
    if(args[i][0]  == '-')
    {
      switch(args[i][1])
      {
        case 'i':
          if(use_regex)
          {
            fprintf(stderr, "-i and -r are mutually exclusive");
            return EXIT_FAILURE;
          }
          ignore_case = TRUE;
          args[i][0] = 0; /* Mark as beeing used */
          break;
        case 'r':
          if(ignore_case)
          {
            fprintf(stderr, "-i and -r are mutually exclusive");
            return EXIT_FAILURE;
          }
          if(args[i][2] != 0)
          {
            if(!reg_exp.compile(&args[i][2]))
            {
              fprintf(stderr, "The regular expression does not compile\n");
              fprintf(stderr, "%s\n", reg_exp.get_last_error());
              return EXIT_FAILURE;
            }
            got_pattern = TRUE;
          }
          use_regex = TRUE;
          args[i][0] = 0; /* Mark as beeing used */
          break;
        #ifndef __GNUC__
        case 's': /* Was to lazy to implement this on GCC */
          parse_subdir = TRUE;
          args[i][0] = 0; /* Mark as beeing used */
          break;
        #endif
        case 'h':
        case '?':
          show_help = TRUE;
          break;
        default:
          break;
      }
    }
  }
  if(show_help)
  {
    help(use_regex);
    return EXIT_FAILURE;
  }

  for(i = 1; i < argc; i++)
  {
    if(args[i][0] != 0)
    {
      if(use_regex && !got_pattern)
      {
        if(!reg_exp.compile(args[i]))
        {
          fprintf(stderr, "The regular expression does not compile\n");
          fprintf(stderr, "%s\n", reg_exp.get_last_error());
          return EXIT_FAILURE;
        }
        got_pattern = TRUE;
      }
      else if(pattern[0] == 0 && !use_regex)
      {
        strcpy(pattern, args[i]);
        if(ignore_case)
          for(n = 0; pattern[n]; n++)
            pattern[n] = tolower(pattern[n]);
      }
      else if(strcmp(args[i], "stdin"))
      {
        #ifndef __GNUC__
        char cwd[_MAX_PATH];
        cwd[0] = 0;
        parse_path(args[i], cwd);
        expand(args[i], cwd);
        #else
        grep(args[i]);
        #endif
      }
      else
      {
        char fn[1024], *s;
        while(fgets(fn, sizeof(fn), stdin))
        {
          if((s = strchr(fn, '\n')) != NULL)
            *s = 0;
          grep(fn);
        }
      }
    }
  }
  if(!read_a_file)
  {
    fputs("No files found!\n", stdout);
    help(FALSE);
    return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;
}

void grep(const char *file)
{
  int line_no = 0, n;
  BOOL hit;
  char *s;
  FILE *f = fopen(file, "rt");

  if(!f)
  {
    fprintf(stderr, "Unable to open: %s\n", file);
    return;
  }
  while(fgets(line, sizeof(line), f))
  {
    line_no++;
    if((s = strchr(line, '\n')) != NULL)
      *s = 0;
    if(use_regex)
      hit = reg_exp.find(line) ? TRUE : FALSE;
    else if(ignore_case)
    {
      for(n = 0; line[n]; n++)
        l_line[n] = tolower(line[n]);
      l_line[n] = 0;
      hit = strstr(l_line, pattern) ? TRUE : FALSE;
    }
    else
      hit = strstr(line, pattern) ? TRUE : FALSE;
    if(hit)
      fprintf(stdout, "%s:%d:%s\n", file, line_no, line);
  }
  read_a_file = TRUE;
  fclose(f);
}

const char * const help_message[] =
{
#ifdef __GNUC__
  "Usage: wgrep [-i|-r] pattern files\n",
#else
  "Usage: wgrep [-i|-r] [-s] pattern files\n",
#endif
  "-i Ignore case\n",
  "-r Regular expression search\n",
  "-h Print this help message and exit, use -h -r to display regexp help\n",
#ifndef __GNUC__
  "-s Search Subdirectories\n",
#endif
  "pattern is the pattern to search for or the regular expression\n",
  "If the pattern contains spaces enclose it in double quotes\n",
#ifdef __GNUC__
  "If the pattern includes * or ? or begins with a @ use double quotes\n",
#endif
  "Files is the list of files to search for\n",
  "The files to search can be read from a file if specified as @file.lst\n",
  "Wildcards (* and ?) are accepted\n",
  "If files is specifed as stdin, stdin will be read for files to scan\n",
  "This is usefull as you can pipe from another command eg:\n",
  "dir /B /S *.c |wgrep foo stdin\n",
  "This will search all c-files in current directory and sub directoreis\n",
   NULL
};

const char * const regexp_help_message[] =
{
  "A regular expression is zero or more branches, separated by `|'.\n",
  "It matches anything that matches one of the branches.\n",
  "A branch is zero or more pieces, concatenated. It matches\n",
  "a match for the first, followed by a match for the second, etc.\n",
  "A piece is an atom possibly followed by `*', `+', or `?'.\n",
  "An atom followed by `*' matches a sequence of 0 or more matches of the atom.\n",
  "An atom followed by `+' matches a sequence of 1 or more matches of the atom.\n",
  "An atom followed by `?' matches a match of the atom, or the null string.\n",
  "An atom is a regular expression in parentheses (matching a match for the\n",
  "regular expression), a range (see below), `.' (matching any single character),\n",
  "`^' (matching the null string at the beginning of the input string),\n",
  "`$' (matching the null string at the end of the input string),\n",
  "a `\\' followed by a single character (matching that character),\n",
  "or a single character with no other significance (matching that character).\n",
  "A range is a sequence of characters enclosed in `[]'. It normally matches\n",
  "any single character from the sequence. If the sequence begins with\n",
  "`^', it matches any single character not from the rest of the sequence.\n",
  "If two characters in the sequence are separated by `-', this is shorthand\n",
  "for the full list of ASCII characters between them (e.g. `[0-9]' matches\n",
  "any decimal digit). To search for -]^\\ in a range, specify \\-, \\], \\^, \\\\\n",
  "to search for tab, use \\t\n",
  "\n",
  "There is a copyright on the regular expression libary:\n",
  "Copyright (c) 1986 by University of Toronto.\n",
  "Written by Henry Spencer. Not derived from licensed software.\n",
  "See regexp.cpp for more info\n",
  "There is no restriction for wgrep itself, other than those arising from\n"
  "the copyright of the regexp. libary\n",
  NULL
};

void help(BOOL regex)
{
  int i;
  for(i = 0; help_message[i]; i++)
    fputs(help_message[i], stdout);
  if(regex)
    for(i = 0; regexp_help_message[i]; i++)
      fputs(regexp_help_message[i], stdout);
}

#ifndef __GNUC__
/* Borland & Microsoft needs help to expand wildcards and to handle file lists */
void expand(char *file, char *path)
{
  if(file[0] == '@')
  {
    FILE *f = fopen(&file[1], "rt");
    if(f)
    {
      char line[1024], *s;
      while(fgets(line, sizeof(line), f))
      {
        if((s = strchr(line, '\n')) != NULL)
          *s = 0;
        if(strlen(line))
        {
          char cwd[_MAX_PATH];
          parse_path(line, cwd);
          expand(line, cwd);
        }
      }
      fclose(f);
    }
  }
  else
  {
    char temp[_MAX_PATH];
    struct _finddata_t fileinfo;
    if(path[0] && path[strlen(path) - 1] != '\\')
      strcat(path, "\\");
    sprintf(temp, "%s%s", path, file);
    long handle = _findfirst(temp, &fileinfo);
    if(handle != -1)
    {
      /* First handle all files */
      do
      {
        if(!(fileinfo.attrib & _A_SUBDIR))
        {
          char t2[_MAX_PATH];
          sprintf(t2, "%s%s", path, fileinfo.name);
          grep(t2);
        }
      }
      while(_findnext(handle, &fileinfo) == 0);
      _findclose(handle);
    }
    /* Then handle subdir (as requested) */
    if(parse_subdir)
    {
      sprintf(temp, "%s*.*", path);
      handle = _findfirst(temp, &fileinfo);
      if(handle != -1)
      {
        do
        {
          if((fileinfo.attrib & _A_SUBDIR) && strcmp(fileinfo.name, ".") && strcmp(fileinfo.name, ".."))
          {
            char t3[_MAX_PATH];
            sprintf(t3, "%s%s", path, fileinfo.name);
            expand(file, t3);
          }
        }
        while(_findnext(handle, &fileinfo) == 0);
        _findclose(handle);
      }
    }
  }
}

void parse_path(char *file, char *path)
/******************************************************************************
 * Description:
 *   Split the file into path and filename
 *   if parse_subdir then path will be the full path of the file
 * Parameters:
 *   file:
 *   path:
 *****************************************************************************/
{
  char temp[2*_MAX_PATH], *s, *p, cwd[_MAX_PATH];
  char *path_part[_MAX_PATH];
  int num_path_part = 0, i, k;
  char drive[_MAX_PATH], dir[_MAX_PATH], fn[_MAX_PATH], ext[_MAX_PATH];

  /* Let's get rid of and /'s */
  while((s = strchr(file, '/')) != NULL)
    *s = '\\';

  if(*file == '@') /* It's a list file, don't process any futher */
    return;

  if(file[1] == ':' || !parse_subdir)
  { /* Simple case full path is given */
    _splitpath(file, drive, dir, fn, ext);
    strcpy(path, drive);
    strcat(path, dir);
    strcpy(file, fn);
    strcat(file, ext);
    return;
  }
  if(file[0] == '\\')
  { /* From root, simply add the drive */
    getcwd(temp, _MAX_PATH);
    _splitpath(file, drive, dir, fn, ext);
    path[0] = temp[0];
    path[1] = temp[1];
    path[2] = 0;
    strcat(path, dir);
    strcpy(file, fn);
    strcat(file, ext);
    return;
  }

  getcwd(cwd, sizeof(cwd));
  if(strlen(cwd) && cwd[strlen(cwd) - 1] != '\\')
    strcat(cwd, "\\");
  sprintf(temp, "%s%s", cwd, file);

  _splitpath(temp, drive, dir, fn, ext);
  s = strtok(dir, "\\");
  while(s)
  {
    if(strcmp(s, "."))
      path_part[num_path_part++] = s;
    s = strtok(NULL, "\\");
  }
  temp[0] = '\\';
  temp[1] = 0;

  for(i = 0; i < num_path_part; i++)
  {
    k = 0;
    while(i < num_path_part - 1 && !strcmp(path_part[i + 1], ".."))
    {
      i++; /* Skip the next path_part */
      k = 1;
    }
    if(!k && strlen(path_part[i]))
    {
      strcat(temp, path_part[i]);
      strcat(temp, "\\");
    }
  }

  sprintf(path, "%s%s", drive, temp);
  sprintf(file, "%s%s", fn, ext);
}
#endif
